@@ -2,7 +2,7 @@
 qrq - High speed morse trainer, similar to the DOS classic "Rufz"
 Copyright (C) 2006-2011  Fabian Kurz
 
-$Id: qrq.c 360 2011-09-16 11:21:41Z KB1OOO $
+$Id: qrq.c 388 2011-12-18 21:28:50Z dj1yfk $
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -18,9 +18,13 @@ this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
 Street, Fifth Floor, Boston, MA  02110-1301, USA.
 
 */ 
+#if WIN32
+#define WIN_THREADS
+#endif
 
-
+#ifndef WIN_THREADS
 #include <pthread.h>			/* CW output will be in a separate thread */
+#endif
 #include <ncurses.h>
 #include <stdlib.h>
 #include <string.h>
@@ -35,7 +39,9 @@ Street, Fifth Floor, Boston, MA  02110-1301, USA.
 #include <sys/stat.h>			/* mkdir */
 #include <sys/types.h>
 #include <errno.h>
-
+#ifdef WIN32
+#include <windows.h>
+#endif
 
 #define PI M_PI
 
@@ -86,7 +92,7 @@ static char mycall[15]="DJ1YFK";		/* mycall. will be read from qrqrc */
 static char dspdevice[PATH_MAX]="/dev/dsp";	/* will also be read from qrqrc */
 static int score = 0;					/* qrq score */
 static int sending_complete;			/* global lock for "enter" while sending */
-static int callnr;						/* nr of actual call in attempt */
+static int callnr = 0;					/* nr of actual call in attempt */
 static int initialspeed=200;			/* initial speed. to be read from file*/
 static int mincharspeed=0;				/* min. char. speed, below: farnsworth*/
 static int speed=200;					/* current speed in cpm */
@@ -138,10 +144,18 @@ static int read_callbase ();
 static void find_callbases();
 static void select_callbase ();
 static void help ();
+static void callbase_dialog();
+static void parameter_dialog();
+static int clear_parameter_display();
+static void update_parameter_dialog();
 
+#ifdef WIN_THREADS
+HANDLE cwthread;
+#else
 pthread_t cwthread;				/* thread for CW output, to enable
 								   keyboard reading at the same time */
 pthread_attr_t cwattr;
+#endif
 
 char rcfilename[PATH_MAX]="";			/* filename and path to qrqrc */
 char tlfilename[PATH_MAX]="";			/* filename and path to toplist */
@@ -153,9 +167,12 @@ char destdir[PATH_MAX]="";
 /* create windows */
 WINDOW *top_w;					/* actual score					*/
 WINDOW *mid_w;					/* callsign history/mistakes	*/
+WINDOW *conf_w;					/* parameter config display	*/
 WINDOW *bot_w;					/* user input line				*/
+WINDOW *inf_w;					/* info window for param displ	*/
 WINDOW *right_w;				/* highscore list/settings		*/
 
+
 int main (int argc, char *argv[]) {
 
   /* if built as osx bundle set DESTDIR to Resources dir of bundle */
@@ -170,6 +187,7 @@ int main (int argc, char *argv[]) {
   strcpy(destdir, DESTDIR);
 #endif
 
+	char abort = 0;
 	char tmp[80]="";
 	char input[15]="";
 	int i=0,j=0;						/* counter etc. */
@@ -185,7 +203,7 @@ int main (int argc, char *argv[]) {
 	curs_set(FALSE);
 	keypad(stdscr, TRUE);
 	scrollok(stdscr, FALSE);
-
+	
 	printw("qrq v%s - Copyright (C) 2006-2011 Fabian Kurz, DJ1YFK\n", VERSION);
 	printw("This is free software, and you are welcome to redistribute it\n");
 	printw("under certain conditions (see COPYING).\n");
@@ -200,7 +218,6 @@ int main (int argc, char *argv[]) {
 	 * is 31 characters long, with the added time stamp */
 	check_toplist();
 
-
 	/* buffer for audio */
 	for (long_i=0;long_i<88200;long_i++) {
 		buffer[long_i]=0;
@@ -209,11 +226,13 @@ int main (int argc, char *argv[]) {
 	/* random seed from time */
 	srand( (unsigned) time(NULL) ); 
 
+#ifndef WIN_THREADS
 	/* Initialize cwthread. We have to wait for the cwthread to finish before
 	 * the next cw output can be made, this will be done with pthread_join */
 	pthread_attr_init(&cwattr);
 	pthread_attr_setdetachstate(&cwattr, PTHREAD_CREATE_JOINABLE);
-
+#endif
+	
 	/****** Reading configuration file ******/
 	printw("\nReading configuration file qrqrc \n");
 	read_config();
@@ -233,13 +252,33 @@ int main (int argc, char *argv[]) {
 	refresh();
 	getch();
 
+	erase();
+	refresh();
+
 	top_w = newwin(4, 60, 0, 0);
 	mid_w = newwin(17, 60, 4, 0);
+	conf_w = newwin(17, 60, 4, 0);
 	bot_w = newwin(3, 60, 21, 0);
+	inf_w = newwin(3, 60, 21, 0);
 	right_w = newwin(24, 20, 0, 60);
 
+	werase(top_w);
+	werase(mid_w);
+	werase(conf_w);
+	werase(bot_w);
+	werase(inf_w);
+	werase(right_w);
+
+	keypad(bot_w, TRUE);
+	keypad(mid_w, TRUE);
+	keypad(conf_w, TRUE);
+
+#ifdef WIN_THREADS
+	cwthread = (HANDLE) _beginthread( morse,0,"QRQ");
+#else
 	/* no need to join here, this is the first possible time CW is sent */
 	pthread_create(&cwthread, NULL, & morse, (void *) "QRQ");
+#endif
 
 /* very outter loop */
 while (1) {	
@@ -247,8 +286,10 @@ while (1) {
 /* status 1 = running an attempt of 50 calls */	
 while (status == 1) {
 	box(top_w,0,0);
+	box(conf_w,0,0);
 	box(mid_w,0,0);
 	box(bot_w,0,0);
+	box(inf_w,0,0);
 	box(right_w,0,0);
 	wattron(top_w,A_BOLD);
 	mvwaddstr(top_w,1,1, "QRQ v");
@@ -272,7 +313,9 @@ while (status == 1) {
 	mvwaddstr(mid_w,8,2, "errors is credited.");
 	mvwaddstr(mid_w,10,2, "F6 repeats a callsign once, F10 quits.");
 	mvwaddstr(mid_w,12,2, "Settings can be changed with F5 (or in qrqrc).");
+#ifndef WIN32
 	mvwaddstr(mid_w,14,2, "Score statistics (requires gnuplot) with F7.");
+#endif
 
 	wattron(right_w,A_BOLD);
 	mvwaddstr(right_w,1, 6, "Toplist");
@@ -289,29 +332,36 @@ while (status == 1) {
 	wrefresh(mid_w);
 	wrefresh(bot_w);
 	wrefresh(right_w); 
-
+	
 	/* reset */
 	maxspeed = errornr = score = 0;
 	speed = initialspeed;
 	
 	/* prompt for own callsign */
-	i = readline(bot_w, 1, 30, mycall, 0);
+	i = readline(bot_w, 1, 30, mycall, 1);
 
 	/* F5 -> Configure sound */
 	if (i == 5) {
-		status = 2;
+		parameter_dialog();
 		break;
 	} 
 	/* F6 -> play test CW */
 	else if (i == 6) {
 		freq = constanttone ? ctonefreq : 800;
+#ifdef WIN_THREADS
+		 WaitForSingleObject(cwthread,INFINITE);
+		 cwthread = (HANDLE) _beginthread( morse,0,"VVVTEST");
+#else
 		pthread_join(cwthread, NULL);
 		j = pthread_create(&cwthread, NULL, &morse, (void *) "VVVTEST");	
 		thread_fail(j);
+#endif
 		break;
 	}
 	else if (i == 7) {
+#ifndef WIN32
 		statistics();
+#endif
 		break;
 	}
 
@@ -346,8 +396,11 @@ while (status == 1) {
 	for (callnr=1; callnr < (unlimitedattempt ? nrofcalls : 51); callnr++) {
 		/* Make sure to wait for the cwthread of the previous callsign, if
 		 * necessary. */
+#ifdef WIN_THREADS
+		WaitForSingleObject(cwthread,INFINITE);
+#else
 		pthread_join(cwthread, NULL);
-		
+#endif	
 		/* select an unused callsign from the calls-array */
 		do {
 			i = (int) ((float) nrofcalls*rand()/(RAND_MAX+1.0));
@@ -378,20 +431,50 @@ while (status == 1) {
 		 * input and echoing at the same time possible */
 		
 		sending_complete = 0;	
+#ifdef WIN_THREADS
+		cwthread = (HANDLE) _beginthread( morse,0,calls[i]);
+#else
 		j = pthread_create(&cwthread, NULL, morse, calls[i]);	
 		thread_fail(j);		
+#endif
 		
 		f6pressed=0;
-		while (readline(bot_w, 1, 8, input,1) > 4) {/* F5 or F6 was pressed */
-			if (f6pressed && (f6 == 0)) {
-				continue;
-			}
-			f6pressed=1;
-			/* wait for old cwthread to finish, then send call again */
+
+	
+		while (!abort && (j = readline(bot_w, 1, 8, input,1)) > 4) {/* F5..F10 pressed */
+
+			switch (j) {
+				case 6:
+				if (f6pressed && (f6 == 0)) {
+					continue;
+				}
+				f6pressed=1;
+				/* wait for old cwthread to finish, then send call again */
+			
+#ifdef WIN_THREADS
+			WaitForSingleObject(cwthread,INFINITE);
+			cwthread = (HANDLE) _beginthread( morse,0,calls[i]);
+#else
 			pthread_join(cwthread, NULL);
-			j = pthread_create(&cwthread, NULL, morse, calls[i]);	
-			thread_fail(j);		
+			j = pthread_create(&cwthread, NULL, &morse, calls[i]);	
+			thread_fail(j);
+#endif	
+					break; /* 6*/
+				case 10:	/* abort attempt */
+					abort = 1;
+					continue;
+					break;
+			}
+					
 		}
+
+		
+		if (abort) {
+			abort = 0;
+			input[0]='\0';
+			break;
+		}
+		
 		tmp[0]='\0';	
 		score += calc_score(calls[i], input, speed, tmp);
 		update_score();
@@ -403,77 +486,56 @@ while (status == 1) {
 	}
 
 	/* attempt is over, send AR */
-	pthread_join(cwthread, NULL);
-	j = pthread_create(&cwthread, NULL, &morse, (void *) "+");	
-
+	callnr = 0;
+	
+#ifdef WIN_THREADS
+		 WaitForSingleObject(cwthread,INFINITE);
+		 cwthread = (HANDLE) _beginthread( morse,0,"+");
+#else
+		pthread_join(cwthread, NULL);
+		j = pthread_create(&cwthread, NULL, &morse, (void *) "+");	
+		thread_fail(j);
+#endif
+	
 	add_to_toplist(mycall, score, maxspeed);
 	
+	curs_set(0);
 	wattron(bot_w,A_BOLD);
 	mvwprintw(bot_w,1,1, "Attempt finished. Press any key to continue!");
 	wattroff(bot_w,A_BOLD);
 	wrefresh(bot_w);
 	getch();
 	mvwprintw(bot_w,1,1, "                                            ");
+	curs_set(1);
+
 	
 } /* while (status == 1) */
 
-/* status == 2. Change parameters */
-while (status == 2) {
-	clear_display();
 
-	switch (waveform) {
-		case SINE:
-			strcpy(wavename, "Sine    ");
-			break;
-		case SAWTOOTH:
-			strcpy(wavename, "Sawtooth");
-			break;
-		case SQUARE:
-			strcpy(wavename, "Square  ");
-			break;
-	}
+} /* very outter loop */
 
-	mvwaddstr(bot_w,1,1, "                                                         ");
-	curs_set(0);
-	wattron(mid_w,A_BOLD);
-	mvwaddstr(mid_w,1,1, "Configuration:          Value                Change");
-	mvwprintw(mid_w,14,2, "      F6                    F10            ");
-	mvwprintw(mid_w,15,2, "      F2");
-	wattroff(mid_w, A_BOLD);
-	mvwprintw(mid_w,2,2, "Initial Speed:         %3d CpM / %3d WpM" 
-					"    up/down", initialspeed, initialspeed/5);
-	mvwprintw(mid_w,3,2, "Min. character Speed:  %3d CpM / %3d WpM" 
-					"    left/right", mincharspeed, mincharspeed/5);
-	mvwprintw(mid_w,4,2, "CW rise/falltime (ms): %1.1f           " 
-					"       +/-", edge);
-	mvwprintw(mid_w,5,2, "Callsign:              %-14s" 
-					"       c", mycall);
-	mvwprintw(mid_w,6,2, "CW pitch (0 = random): %-4d"
-					"                 k/l or 0", (constanttone)?ctonefreq : 0);
-	mvwprintw(mid_w,7,2, "CW waveform:           %-8s"
-					"             w", wavename);
-	mvwprintw(mid_w,8,2, "Allow unlimited F6*:   %-3s"
-					"                  f", (f6 ? "yes" : "no"));
-	mvwprintw(mid_w,9,2, "Fixed CW speed*:       %-3s"
-					"                  s", (fixspeed ? "yes" : "no"));
-	mvwprintw(mid_w,10,2, "Unlimited attempt*:    %-3s"
-					"                  u", (unlimitedattempt ? "yes" : "no"));
-	mvwprintw(mid_w,11,2, "Callsign database:     %-15s"
-					"      d (%d)", basename(cbfilename),nrofcalls);
-#ifdef OSS
-	mvwprintw(mid_w,12,2, "DSP device:            %-15s"
-					"      e", dspdevice);
-#endif
-	mvwprintw(mid_w,14,2, "Press");
-	mvwprintw(mid_w,14,11, "to play sample CW,");
-	mvwprintw(mid_w,14,34, "to go back.");
-	mvwprintw(mid_w,15,2, "Press");
-	mvwprintw(mid_w,15,11, "to save config permanently.");
-	mvwprintw(bot_w,1,11, "* Makes scores ineligible for toplist");
-	wrefresh(mid_w);
-	wrefresh(bot_w);
-	
-	j = getch();
+	getch();
+	endwin();
+	delwin(top_w);
+	delwin(bot_w);
+	delwin(mid_w);
+	delwin(right_w);
+	getch();
+	return 0;
+}
+
+
+
+/* (formerly status == 2). Change parameters */
+
+void parameter_dialog () {
+
+int j = 0;
+
+
+update_parameter_dialog();
+
+while ((j = getch()) != 0) {
 
 	switch ((int) j) {
 		case '+':							/* rise/falltime */
@@ -539,7 +601,7 @@ while (status == 2) {
 			}
 			break;
 		case 'c':
-			readline(mid_w, 5, 25, mycall, 1);
+			readline(conf_w, 5, 25, mycall, 1);
 			if (strlen(mycall) == 0) {
 				strcpy(mycall, "NOCALL");
 			}
@@ -550,7 +612,7 @@ while (status == 2) {
 			break;
 #ifdef OSS
 		case 'e':
-			readline(mid_w, 12, 25, dspdevice, 0);
+			readline(conf_w, 12, 25, dspdevice, 0);
 			if (strlen(dspdevice) == 0) {
 				strcpy(dspdevice, "/dev/dsp");
 			}
@@ -558,25 +620,43 @@ while (status == 2) {
 			break;
 #endif
 		case 'd':							/* go to database browser */
-				status = 3;
+			if (!callnr) {					/* Only allow outside of attempt */
 				curs_set(1);
+				callbase_dialog();
+			}
 			break;
 		case KEY_F(2):
 			save_config();	
-			mvwprintw(mid_w,15,39, "Config saved!");
-			wrefresh(mid_w);
+			mvwprintw(conf_w,15,39, "Config saved!");
+			wrefresh(conf_w);
+#ifdef WIN32
+			Sleep(1000);
+#else
 			sleep(1);	
+#endif
 			break;
 		case KEY_F(6):
 			freq = constanttone ? ctonefreq : 800;
-			pthread_join(cwthread, NULL);
-			j = pthread_create(&cwthread, NULL, &morse, (void *) "TESTING");	
-			thread_fail(j);
+#ifdef WIN_THREADS
+		 WaitForSingleObject(cwthread,INFINITE);
+		 cwthread = (HANDLE) _beginthread( morse,0,"TESTING");
+#else
+		pthread_join(cwthread, NULL);
+		j = pthread_create(&cwthread, NULL, &morse, (void *) "TESTING");	
+		thread_fail(j);
+#endif
 			break;
 		case KEY_F(10):
 		case KEY_F(3):
-			status = 1;
 			curs_set(1);
+			clear_parameter_display();
+			wrefresh(conf_w);
+			/* restore old windows */
+			touchwin(mid_w);
+			touchwin(bot_w);
+			wrefresh(mid_w);
+			wrefresh(bot_w);
+			return;
 	}
 
 	speed = initialspeed;
@@ -585,25 +665,104 @@ while (status == 2) {
 	if (f6 || fixspeed || unlimitedattempt) {
 		attemptvalid = 0;	
 	}
-}
 
-while (status == 3) {
+	update_parameter_dialog();
 
-	clear_display();
+} /* while 1 (return only by F3/F10) */
 
-	wattron(mid_w,A_BOLD);
-	mvwaddstr(mid_w,1,1, "Change Callsign Database");
-	wattroff(mid_w,A_BOLD);
-	mvwprintw(mid_w,3,1, ".qcb files found (in %s/share/qrq/ and ~/.qrq/):",destdir);
-/*	mvwaddstr(mid_w,3,1, ".qcb files found:");*/ 
+} /* parameter_dialog */
+
+
+/* update_parameter_dialog 
+ * repaints the whole config/parameter screen (F5) */
+
+
+void update_parameter_dialog () {
+
+	clear_parameter_display();
+	switch (waveform) {
+		case SINE:
+			strcpy(wavename, "Sine    ");
+			break;
+		case SAWTOOTH:
+			strcpy(wavename, "Sawtooth");
+			break;
+		case SQUARE:
+			strcpy(wavename, "Square  ");
+			break;
+	}
+
+	mvwaddstr(inf_w,1,1, "                                                         ");
+	curs_set(0);
+	wattron(conf_w,A_BOLD);
+	mvwaddstr(conf_w,1,1, "Configuration:          Value                Change");
+	mvwprintw(conf_w,14,2, "      F6                    F10            ");
+	mvwprintw(conf_w,15,2, "      F2");
+	wattroff(conf_w, A_BOLD);
+	mvwprintw(conf_w,2,2, "Initial Speed:         %3d CpM / %3d WpM" 
+					"    up/down", initialspeed, initialspeed/5);
+	mvwprintw(conf_w,3,2, "Min. character Speed:  %3d CpM / %3d WpM" 
+					"    left/right", mincharspeed, mincharspeed/5);
+	mvwprintw(conf_w,4,2, "CW rise/falltime (ms): %1.1f           " 
+					"       +/-", edge);
+	mvwprintw(conf_w,5,2, "Callsign:              %-14s" 
+					"       c", mycall);
+	mvwprintw(conf_w,6,2, "CW pitch (0 = random): %-4d"
+					"                 k/l or 0", (constanttone)?ctonefreq : 0);
+	mvwprintw(conf_w,7,2, "CW waveform:           %-8s"
+					"             w", wavename);
+	mvwprintw(conf_w,8,2, "Allow unlimited F6*:   %-3s"
+					"                  f", (f6 ? "yes" : "no"));
+	mvwprintw(conf_w,9,2, "Fixed CW speed*:       %-3s"
+					"                  s", (fixspeed ? "yes" : "no"));
+	mvwprintw(conf_w,10,2, "Unlimited attempt*:    %-3s"
+					"                  u", (unlimitedattempt ? "yes" : "no"));
+	if (!callnr) {
+		mvwprintw(conf_w,11,2, "Callsign database:     %-15s"
+					"      d (%d)", basename(cbfilename),nrofcalls);
+	}
+#ifdef OSS
+	mvwprintw(conf_w,12,2, "DSP device:            %-15s"
+					"      e", dspdevice);
+#endif
+	mvwprintw(conf_w,14,2, "Press");
+	mvwprintw(conf_w,14,11, "to play sample CW,");
+	mvwprintw(conf_w,14,34, "to go back.");
+	mvwprintw(conf_w,15,2, "Press");
+	mvwprintw(conf_w,15,11, "to save config permanently.");
+	mvwprintw(inf_w,1,1, "          * Makes scores ineligible for toplist");
+	wrefresh(conf_w);
+	wrefresh(inf_w);
+	
+
+} /* update_parameter_dialog */
+
+
+
+
+void callbase_dialog () {
+
+	clear_parameter_display();
+
+	wattron(conf_w,A_BOLD);
+	mvwaddstr(conf_w,1,1, "Change Callsign Database");
+	wattroff(conf_w,A_BOLD);
+#if WIN32
+	mvwprintw(conf_w,3,1, ".qcb files found:");
+#else
+	mvwprintw(conf_w,3,1, ".qcb files found (in %s/share/qrq/ and ~/.qrq/):",destdir);
+#endif
 
 	/* populate cblist */	
 	find_callbases();
 	/* selection dialog */
 	select_callbase();
-	
-	wrefresh(mid_w);
-	status = 2;	/* back to config menu */
+	wrefresh(conf_w);
+
+	nrofcalls = read_callbase();
+
+
+	return;	/* back to config menu */
 }
 
 
@@ -617,17 +776,6 @@ while (status == 3) {
 
 
 
-} /* very outter loop */
-
-	getch();
-	endwin();
-	delwin(top_w);
-	delwin(bot_w);
-	delwin(mid_w);
-	delwin(right_w);
-	getch();
-	return 0;
-}
 
 
 /* reads a callsign etc. in *win at y/x and writes it to *line */
@@ -651,8 +799,7 @@ static int readline(WINDOW *win, int y, int x, char *line, int capitals) {
 	curs_set(TRUE);
 	
 	while (1) {
-		c = getch();
-
+		c = wgetch(win);
 		if (c == '\n' && sending_complete)
 			break;
 
@@ -671,7 +818,7 @@ static int readline(WINDOW *win, int y, int x, char *line, int capitals) {
 			line[p]=c;						/* insert into gap */
 			p++;
 		}
-		else if ((c == KEY_BACKSPACE || c == 127 || c == 9)
+		else if ((c == KEY_BACKSPACE || c == 127 || c == 9 || c == 8)
 						&& p != 0) {					/* BACKSPACE */
 			for (i=p-1;i < strlen(line); i++) {
 				line[i] =  line[i+1];
@@ -718,7 +865,7 @@ static int readline(WINDOW *win, int y, int x, char *line, int capitals) {
 			wrefresh(top_w);
 		}
 		else if (c == KEY_F(5)) {
-			return 5;
+			parameter_dialog();
 		}
 		else if (c == KEY_F(6)) {
 			return 6;
@@ -727,17 +874,27 @@ static int readline(WINDOW *win, int y, int x, char *line, int capitals) {
 			return 7;
 		}
 		else if (c == KEY_F(10)) {				/* quit */
+			if (callnr) {						/* quit attempt only */
+				return 10;
+			} 
+			/* else: quit program */
 			endwin();
 			printf("Thanks for using 'qrq'!\nYou can submit your"
 					" highscore to http://fkurz.net/ham/qrqtop.php\n");
 			/* make sure that no more output is running, then send 73 & quit */
-			pthread_join(cwthread, NULL);
 			speed = 200; freq = 800;
-			j = pthread_create(&cwthread, NULL, &morse, (void *) "73");	
-			thread_fail(j);
+#ifdef WIN_THREADS
+		 WaitForSingleObject(cwthread,INFINITE);
+		 cwthread = (HANDLE) _beginthread( morse,0,"73");
+		 WaitForSingleObject(cwthread,INFINITE);
+#else
+		pthread_join(cwthread, NULL);
+		j = pthread_create(&cwthread, NULL, &morse, (void *) "73");	
+		thread_fail(j);
 			/* make sure the cw thread doesn't die with the main thread */
-			pthread_join(cwthread, NULL);
 			/* Exit the whole main thread */
+			pthread_join(cwthread, NULL);
+#endif
 			exit(0);
 		}
 		
@@ -878,18 +1035,31 @@ static int clear_display() {
 	return 0;
 }
 
+/* clear parameter display */
+static int clear_parameter_display() {
+	int i;
+	for (i=1;i<16;i++) {
+		mvwprintw(conf_w,i,1,"                                 "
+										"                        ");
+	}
+	return 0;
+}
+
+
 /* write entry into toplist at the right place 
  * going down from the top of the list until the score in the current line is
  * lower than the score made. then */
+
 static int add_to_toplist(char * mycall, int score, int maxspeed) {
 	FILE *fh;	
+	char *part1, *part2;
 	char tmp[35]="";
-	char line[35]="";
 	char insertline[35]="DJ1YFK     36666 333 1111111111";		/* example */
 						/* call       pts   max timestamp */
-	int i=0;
+	int i=0, k=0, j=0;
 	int pos = 0;		/* position where first score < our score appears */
 	int timestamp = 0;
+	int len = 32;		/* length of a line. 32 for unix, 33 for Win */
 
 	/* For the training modes */
 	if (score == 0) {
@@ -897,35 +1067,59 @@ static int add_to_toplist(char * mycall, int score, int maxspeed) {
 	}
 
 	timestamp = (int) time(NULL);
+	sprintf(insertline, "%-10s%6d %3d %10d", mycall, score, maxspeed, timestamp);
 	
-	/* assemble scoreline to insert */
-	sprintf(insertline, "%-10s%6d %3d %10d",mycall, score, maxspeed, timestamp);
-	
-	if ((fh = fopen(tlfilename, "r+")) == NULL) {
-		endwin();
-		perror("Unable to open toplist file 'toplist'!\n");
+	if ((fh = fopen("toplist", "rb+")) == NULL) {
+		printf("Unable to open toplist file 'toplist'!\n");
 		exit(EXIT_FAILURE);
 	}
-	while ((feof(fh) == 0) && (fgets(line, 35, fh) != NULL)) {
+
+	/* find out if we use CRLF or just LF */
+	fgets(tmp, 35, fh);
+	if (tmp[31] == '\r') {	/* CRLF */
+		len = 33;
+		strcat(insertline, "\r\n");
+	}
+	else {
+		len = 32;
+		strcat(insertline, "\n");
+	}
+
+	fseek(fh, 0, SEEK_END);
+	j = ftell(fh);
+
+	part1 = malloc((size_t) j);
+	part2 = malloc((size_t) j + len);	/* one additional entry */
+
+	rewind(fh);
+
+	/* read whole toplist */
+	fread(part1, sizeof(char), (size_t) j, fh);
+
+	/* find first score below "score"; scores at positions 10 + (i*len) */
+
+	do {
+		for (i = 0 ; i < 6 ; i++) {	
+			tmp[i] = part1[i + (10 + pos*len)];
+		}
+		k = atoi(tmp);
 		pos++;
-		for (i=10;i<16;i++) {			/* extract the score to tmp*/
-			tmp[i-10] = line[i];
-		}
-		tmp[i] = '\0';
-		i = atoi(tmp);				/* i = score of current line */
-		if (i < score){ 			/* insert score and shift lines below */
-			score = i;				/* (ugly) */
-			fseek(fh, -32L, SEEK_CUR);
-			fputs(insertline, fh);
-			strcpy(insertline, line);	/* actual line -> print one later */
-		}
-	}
-	fputs(insertline, fh);
-	if (score != i) {					/* last place. add newline! */
-		fputs("\n",fh);
-	}
-	fclose(fh);	
+	} while (score < k);
+ 
+	/* Found it! Insert own score here! */
+	memcpy(part2, part1, len * (pos-1));
+	memcpy(part2 + len * (pos - 1), insertline, len);
+	memcpy(part2 + len * pos , part1 + len * (pos -1), j - len * (pos - 1));
+
+	rewind(fh);
+	fwrite(part2, sizeof(char), (size_t) j + len, fh);
+	fclose(fh);
+
+	free(part1);
+	free(part2);
+
 	return 0;
+
 }
 
 
@@ -947,10 +1141,12 @@ static int read_config () {
 		fprintf(stderr, "Unable to open config file %s!\n", rcfilename);
 		exit(EXIT_FAILURE);
 	}
+
 	while ((feof(fh) == 0) && (fgets(tmp, 80, fh) != NULL)) {
 		i=0;
 		line++;
 		tmp[strlen(tmp)-1]='\0';
+
 		/* find callsign, speed etc. 
 		 * only allow if the lines are beginning at zero, so stuff can be
 		 * commented out easily; return value if strstr must point to tmp*/
@@ -1083,7 +1279,6 @@ static int read_config () {
 			printw("  line  %2d: unlim. att.:  %s\n", line, (unlimitedattempt ? "yes":"no"));
         }
 		else if (tmp == strstr(tmp,"callbase=")) {
-			fprintf(stderr, "%d\n", i);
 			while (isgraph(tmp[i] = tmp[9+i])) {
 				i++;
 			}
@@ -1110,9 +1305,26 @@ static void *morse(void *arg) {
 	int c, fulldotlen, dotlen, dashlen, charspeed, farnsworth, fwdotlen;
 	const char *code;
 
+#if WIN32 /* WinMM simple support by Lukasz Komsta, SP8QED */
+	HWAVEOUT		h;
+	WAVEFORMATEX	wf;
+	WAVEHDR			wh;
+	HANDLE			d;
+
+	wf.wFormatTag = WAVE_FORMAT_PCM;
+	wf.nChannels = 1;
+	wf.wBitsPerSample = 16;
+	wf.nSamplesPerSec = samplerate * 2;
+	wf.nBlockAlign = wf.nChannels * wf.wBitsPerSample / 8;
+	wf.nAvgBytesPerSec = wf.nSamplesPerSec * wf.nBlockAlign;
+	wf.cbSize = 0;
+	d = CreateEvent(0, FALSE, FALSE, 0);
+	if(waveOutOpen(&h, 0, &wf, (DWORD) d, 0, CALLBACK_EVENT) != MMSYSERR_NOERROR);
+
+#else
 	/* opening the DSP device */
 	dsp_fd = open_dsp(dspdevice);
-
+#endif
 	/* set bufpos to 0 */
 
 	full_bufpos = 0; 
@@ -1187,21 +1399,28 @@ static void *morse(void *arg) {
 		}
 	}
 
-//#ifndef PA
-//	write_audio(dsp_fd, buffer, 88200);
-//#else
-//	write_audio(dsp_fd, (int *) buffer, 44100);
-//#endif
 
 #if !defined(PA) && !defined(CA)
 	add_to_buf(buffer, 88200);
 #endif
 
+#if WIN32
+	wh.lpData = (char*) &full_buf[0];
+	wh.dwBufferLength = full_bufpos - 2;
+	wh.dwFlags = 0;
+	wh.dwLoops = 0;
+	waveOutPrepareHeader(h, &wh, sizeof(wh));
+	ResetEvent(d);
+	waveOutWrite(h, &wh, sizeof(wh));
+	if(WaitForSingleObject(d, INFINITE) != WAIT_OBJECT_0);
+	waveOutUnprepareHeader(h, &wh, sizeof(wh));
+	waveOutClose(h);
+	CloseHandle(d);
+#else
 	write_audio(dsp_fd, &full_buf[0], full_bufpos);
 	close_audio(dsp_fd);
-
+#endif
 	sending_complete = 1;
-
 	return NULL;
 }
 
@@ -1247,83 +1466,155 @@ static int tonegen (int freq, int len, int waveform) {
 		out = out + (out<<16);	/* stereo only for OSS & CoreAudio*/
 #endif
 		add_to_buf(&out, sizeof(out));
-//		write_audio(dsp_fd, &out, sizeof(out));
 	}
 	return 0;
 }
 
-/* TODO: Remove copypasta, write small functions to do it */
+/* Save config file
+ *
+ * Tries to keep the old format (including comments, etc.) and adds
+ * config options that were not used yet in the file to the end 
+ * */
 
 static int save_config () {
 	FILE *fh;
 	char tmp[80]="";
-	int i;
-	
-	if ((fh = fopen(rcfilename, "r+")) == NULL) {
+	char confopts[12][80] = {
+		"\ncallsign=", 
+		"\ncallbase=",
+		"\ndspdevice=", 
+		"\ninitialspeed=", 
+		"\nmincharspeed=",
+		"\nwaveform=", 
+		"\nconstanttone=",
+		"\nctonefreq=", 
+		"\nfixspeed=", 
+		"\nunlimitedattempt=", 
+		"\nf6=", 
+		"\nrisetime=" 
+	};
+	char *conf1;
+	char *conf2;
+	char *find, *findend;
+	int i, len, conf1len, conf2len;
+	int j;
+
+	conf2 = malloc(1);
+
+	if ((fh = fopen(rcfilename, "rb")) == NULL) {
 		endwin();
 		fprintf(stderr, "Unable to open config file '%s'!\n", rcfilename);
 		exit(EXIT_FAILURE);
 	}
 
-	while ((feof(fh) == 0) && (fgets(tmp, 80, fh) != NULL)) {
-		tmp[strlen(tmp)-1]='\0';
-		i = strlen(tmp);
-		if (strstr(tmp,"initialspeed=")) {
-			fseek(fh, -(i+1), SEEK_CUR);	/* go to beginning of the line */
-			snprintf(tmp, i+1, "initialspeed=%d ", initialspeed);
-			fputs(tmp, fh);	
+	fseek(fh, 0, SEEK_END);
+	j = (int) ftell(fh);
+	conf1 = malloc((size_t) j+1);
+
+	rewind(fh);
+	i = fread(conf1, sizeof(char), (size_t) j, fh);
+	conf1[j] = '\0';
+	conf1len = j;
+
+	fclose(fh);
+
+	/* The whole config file is now in conf1 
+	 *
+	 * For each config option, search&replace it with the current value.
+	 * Only accept key=value pairs if the key starts on pos 0 of the line
+	 * */
+
+	//endwin();
+	for (i = 0; i < 12; i++) {
+		/* assemble new string for this conf option*/
+		switch (i) {
+			case 0:
+				sprintf(tmp, "%s%s ", confopts[i], mycall);
+				break;
+			case 1:
+				sprintf(tmp, "%s%s ", confopts[i], cbfilename);
+				break;
+			case 2:
+				sprintf(tmp, "%s%s ", confopts[i], dspdevice);
+				break;
+			case 3:
+				sprintf(tmp, "%s%d ", confopts[i], initialspeed);
+				break;
+			case 4:
+				sprintf(tmp, "%s%d ", confopts[i], mincharspeed);
+				break;
+			case 5:
+				sprintf(tmp, "%s%d ", confopts[i], waveform);
+				break;
+			case 6:
+				sprintf(tmp, "%s%d ", confopts[i], constanttone);
+				break;
+			case 7:
+				sprintf(tmp, "%s%d ", confopts[i], ctonefreq);
+				break;
+			case 8:
+				sprintf(tmp, "%s%d ", confopts[i], fixspeed);
+				break;
+			case 9:
+				sprintf(tmp, "%s%d ", confopts[i], unlimitedattempt);
+				break;
+			case 10:
+				sprintf(tmp, "%s%d ", confopts[i], f6);
+				break;
+			case 11:
+				sprintf(tmp, "%s%f ", confopts[i], edge);
+				break;
+		}	
+
+		/* Conf option already in rc-file? */
+		if ((find = strstr(conf1, confopts[i])) != NULL) {
+			/* determine length. */
+			findend = find;
+			findend++;	/* starts with \n, always skip it */
+			while (!isspace(*findend++));
+			len = findend - find;
+			
+			/* old size of conf1: conf1len (see above) 
+			 * new size: conf1len - len + strlen(tmp)*/
+			conf2len = conf1len - len + strlen(tmp);
+			conf2 = realloc(conf2, (size_t) conf2len);
+			memcpy(conf2, conf1, (find - conf1));
+			memcpy(conf2 + (find - conf1), tmp, strlen(tmp));
+			memcpy(conf2 + (find - conf1) + strlen(tmp), findend, 
+					(size_t) conf2len - (find - conf1) - strlen(tmp));
 		}
-		else if (strstr(tmp,"constanttone=")) {
-			fseek(fh, -(i+1), SEEK_CUR);
-			snprintf(tmp, i+1, "constanttone=%d ", constanttone);
-			fputs(tmp, fh);	
+		/* otherwise, add to the end */
+		else {
+			/* CR LF or LF only? */
+			if (strstr(conf1, "\r")) {
+				strcat(tmp, "\r");
+			}
+			strcat(tmp, "\n");
+			conf2len = conf1len + strlen(tmp) - 1;
+			conf2 = realloc(conf2, conf2len);
+			memcpy(conf2, conf1, conf1len-1);  // excl. \0
+			memcpy(conf2 + conf1len - 1, tmp, strlen(tmp)); // (incl. \0)
 		}
-		else if (strstr(tmp,"mincharspeed=")) {
-			fseek(fh, -(i+1), SEEK_CUR);
-			snprintf(tmp, i+1, "mincharspeed=%d ", mincharspeed);
-			fputs(tmp, fh);	
-		}
-		else if (strstr(tmp,"ctonefreq=")) {
-			fseek(fh, -(i+1), SEEK_CUR);
-			snprintf(tmp, i+1, "ctonefreq=%d ", ctonefreq);
-			fputs(tmp, fh);	
-		}
-		else if (strstr(tmp, "risetime=")) {
-			fseek(fh, -(i+1), SEEK_CUR);
-			snprintf(tmp, i+1, "risetime=%f ", edge);
-			fputs(tmp, fh);	
-		}
-		else if (strstr(tmp,"callsign=")) {
-			fseek(fh, -(i+1), SEEK_CUR);
-			snprintf(tmp, i+1, "callsign=%-7s ", mycall);
-			fputs(tmp, fh);	
-		}
-		else if (strstr(tmp,"dspdevice=")) {
-			fseek(fh, -(i+1), SEEK_CUR);
-			snprintf(tmp, i+1, "dspdevice=%s ", dspdevice);
-			fputs(tmp, fh);	
-		}
-		else if (strstr(tmp,"waveform=")) {
-			fseek(fh, -(i+1), SEEK_CUR);
-			snprintf(tmp, i+1, "waveform=%d ", waveform);
-			fputs(tmp, fh);	
-		}
-		else if (strstr(tmp,"f6=")) {
-			fseek(fh, -(i+1), SEEK_CUR);
-			snprintf(tmp, i+1, "f6=%d ", f6);
-			fputs(tmp, fh);	
-		}
-		else if (strstr(tmp,"fixspeed=")) {
-			fseek(fh, -(i+1), SEEK_CUR);
-			snprintf(tmp, i+1, "fixspeed=%d ", fixspeed);
-			fputs(tmp, fh);	
-		}
-		else if (strstr(tmp,"unlimitedattempt=")) {
-			fseek(fh, -(i+1), SEEK_CUR);
-			snprintf(tmp, i+1, "unlimitedattempt=%d ", unlimitedattempt);
-			fputs(tmp, fh);	
+		conf1 = realloc(conf1, (size_t) conf2len);
+		conf1len = conf2len;
+		if (conf1 == NULL) {
+			exit(0);
 		}
+		memcpy(conf1, conf2, conf1len);
 	}
+
+	if ((fh = fopen(rcfilename, "wb")) == NULL) {
+		endwin();
+		fprintf(stderr, "Unable to open config file '%s'!\n", rcfilename);
+		exit(EXIT_FAILURE);
+	}
+
+	fwrite(conf1, conf1len, sizeof(char), fh); 
+	fclose(fh);
+
+	free(conf1);
+	free(conf2);
+
 	return 0;
 }
 		
@@ -1395,7 +1686,7 @@ static int check_toplist () {
  *    there.
  * 4) Nowhere --> Exit.*/
 static int find_files () {
-
+	
 	FILE *fh;
 	const char *homedir = NULL;
 	char tmp_rcfilename[1024] = "";
@@ -1408,12 +1699,19 @@ static int find_files () {
 		((fh = fopen("toplist", "r")) == NULL) ||
 		((fh = fopen("callbase.qcb", "r")) == NULL)) {
 		
-		homedir = getenv("HOME");
-		
+		if ((homedir = getenv("HOME")) != NULL) {
 		printw("... not found in current directory. Checking "
 						"%s/.qrq/...\n", homedir);
 		refresh();
 		strcat(rcfilename, homedir);
+		}
+		else {
+		printw("... not found in current directory. Checking "
+						"./.qrq/...\n", homedir);
+		refresh();
+		strcat(rcfilename, ".");
+		}
+				
 		strcat(rcfilename, "/.qrq/qrqrc");
 	
 		/* check if there is ~/.qrq/qrqrc. If it's there, it's safe to assume
@@ -1449,7 +1747,11 @@ static int find_files () {
 						" toplist there.\n", destdir, homedir);
 				strcpy(rcfilename, homedir);
 				strcat(rcfilename, "/.qrq/");
+#ifdef WIN32
+				j = mkdir(rcfilename);
+#else
 				j = mkdir(rcfilename,  0777);
+#endif
 				if (j && (errno != EEXIST)) {
 					printw("Failed to create %s! Exit.\n", rcfilename);
 					getch();
@@ -1507,7 +1809,6 @@ static int find_files () {
 	}
 	refresh();
 	fclose(fh);
-
 	return 0;
 }
 
@@ -1641,13 +1942,19 @@ void find_callbases () {
 	char path[3][PATH_MAX];
 	int i=0,j=0,k=0;
 
-	strcpy(path[0], getenv("PWD"));
-	strcat(path[0], "/");
-	strcpy(path[1], getenv("HOME"));
-	strcat(path[1], "/.qrq/");
-	//strcpy(path[2], DESTDIR"/share/qrq/");
-	strcpy(path[2], destdir);
-	strcat(path[2], "/share/qrq/");
+#ifndef WIN32
+		strcpy(path[0], getenv("PWD"));
+		strcat(path[0], "/");
+		strcpy(path[1], getenv("HOME"));
+		strcat(path[1], "/.qrq/");
+		strcpy(path[2], destdir);
+		strcat(path[2], "/share/qrq/");
+#else
+		strcpy(path[0], "./");
+		strcpy(path[1], getenv("APPDATA"));
+		strcat(path[1], "/qrq/");
+		strcpy(path[2], "c:\\");
+#endif
 
 	for (i=0; i < 100; i++) {
 		strcpy(cblist[i], "");
@@ -1688,9 +1995,13 @@ void select_callbase () {
 	while (strcmp(cblist[i], "")) i++;
 
 	if (!i) {
-		mvwprintw(mid_w,10,4, "No qcb-files found!");
-		wrefresh(mid_w);
+		mvwprintw(conf_w,10,4, "No qcb-files found!");
+		wrefresh(conf_w);
+#ifdef WIN32
+		Sleep(1000);
+#else
 		sleep(1);
+#endif
 		return;
 	}
 
@@ -1699,21 +2010,21 @@ void select_callbase () {
 
 	/* cls */
 	for (j = 5; j < 16; j++) {
-			mvwprintw(mid_w,j,2, "                                         ");
+			mvwprintw(conf_w,j,2, "                                         ");
 	}
 
 	/* display 10 files, highlight cursor position */
 	for (j = p*10; j < (p+1)*10; j++) {
 		if (j <= i) {
 				cblist_ptr = cblist[j];
-				mvwprintw(mid_w,5+(j - p*10 ),2, "  %s       ", cblist_ptr);
+				mvwprintw(conf_w,5+(j - p*10 ),2, "  %s       ", cblist_ptr);
 		}
 		if (c == j) {						/* cursor */
-			mvwprintw(mid_w,5+(j - p*10),2, ">");
+			mvwprintw(conf_w,5+(j - p*10),2, ">");
 		}
 	}
 	
-	wrefresh(mid_w);
+	wrefresh(conf_w);
 
 	k = getch();
 
@@ -1737,7 +2048,7 @@ void select_callbase () {
 			break;
 	}
 
-	wrefresh(mid_w);
+	wrefresh(conf_w);
 
 	} /* while 1 */
 
